<?php

namespace Drupal\form_helper_widget\Services;

use Drupal\Component\Utility\Html;
use Drupal\Core\Entity\Display\EntityFormDisplayInterface;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Link;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\field_ui\Form\EntityDisplayFormBase;
use Drupal\form_helper_widget\Form\FormHelperWidgetAddForm;
use Psr\Log\InvalidArgumentException;

/**
 * Class FormHelperWidgetConf.
 *
 * @package Drupal\form_helper_widget\Services
 */
class FormHelperWidgetConf {

  use StringTranslationTrait;

  /**
   * Service name.
   *
   * @const string
   */
  const SERVICE_NAME = 'form_helper_widget.conf';

  /**
   * ID of the field id.
   *
   * @const string
   */
  const FIELD_ID = 'value';

  /**
   * ID of the field description.
   *
   * @const string
   */
  const FIELD_DESCRIPTION = 'description';

  /**
   * ID of the field first id.
   *
   * AS the field_id value can be update by the user, we need to stock the
   * first_id as an id.
   *
   * @const string
   */
  const FIELD_FIRST_ID = 'first_id';

  /**
   * ID of the field third_party_id.
   *
   * @const string
   */
  const THIRD_PARTY_ID = 'form_helper_widget';

  /**
   * Entity Type Manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * The messenger.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * FormHelperWidgetConf constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManager $entityTypeManager
   *   The entityTypeManager.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The messenger.
   */
  public function __construct(EntityTypeManager $entityTypeManager, MessengerInterface $messenger) {
    $this->entityTypeManager = $entityTypeManager;
    $this->messenger = $messenger;
  }

  /**
   * Retourne le singleton.
   *
   * @return static
   *   Le singleton.
   */
  public static function me() {
    return \Drupal::service(static::SERVICE_NAME);
  }

  /**
   * Ajoute le widget dans la conf tierse.
   *
   * @param mixed $group
   *   The group.
   * @param mixed|null $display
   *   THe display.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null|static
   *   The display.
   */
  public function saveWidget($group, $display = NULL) {
    if ($display === NULL) {
      if ($group->context == 'form') {
        $display = EntityFormDisplay::load($group->entity_type . '.' . $group->bundle . '.' . $group->mode);
      }
    }

    // If no display was found. It doesn't exist yet, create it.
    if (!isset($display)) {
      if ($group->context === 'form') {
        $display = EntityFormDisplay::create([
          'targetEntityType' => $group->entity_type,
          'bundle'           => $group->bundle,
          'mode'             => $group->mode,
        ])->setStatus(TRUE);
      }
    }

    if (isset($display)) {
      $data = (array) $group;
      unset($data['entity_type'], $data['bundle'], $data['mode'], $data['form'], $data['context']);
      // On slugify l'id pour la gestion de ponctuation.
      $id = substr(FormHelperWidgetManager::me()
        ->slugify($data[static::FIELD_ID]), 0, 10) . rand(1, 100);
      $data[static::FIELD_FIRST_ID] = $id;
      $display->setThirdPartySetting(static::THIRD_PARTY_ID, $id, $data);
      $display->save();
    }

    return $display;
  }

  /**
   * Init the render of the widget.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The formstate.
   */
  public function showWidgetConf(array &$form, FormStateInterface $formState) {
    // Only start altering the form if we need to.
    if (empty($form['#fields']) && empty($form['#extra'])) {
      return;
    }

    $display = $this->getDisplayFromFormState($formState);
    $params = field_group_field_ui_form_params($form, $display);

    // Récupération des translation_widgets.
    $this->cleanThirdParty($params->entity_type, $params->bundle, $params->mode);
    $widgets = $display->getThirdPartySettings(static::THIRD_PARTY_ID);

    $formStateValues = $formState->getValues();
    foreach ($widgets as $id => $widget) {
      if ($id) {
        // Save the group when the configuration is submitted.
        if ($formState->getTriggeringElement()['#name'] == FormHelperWidgetManager::me()
          ->slugify($id) . '_plugin_settings_update') {
          $this->updateWidget($form, $formState, $id, $formStateValues['fields'][$id]['format']);
        }

        $this->addWidgetConfig($form['fields'], $id, $widget, $formState, $params, $display);
        $form['#' . static::THIRD_PARTY_ID][$id] = $widget;
      }
    }

    // Gestion de la soumission du formulaire.
    array_unshift($form['actions']['submit']['#submit'], get_called_class() . '::onConfigFormSubmit');
  }

  /**
   * Save settings on configuration form submission.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The formstate.
   */
  public static function onConfigFormSubmit(array &$form, FormStateInterface $formState) {
    if (array_key_exists('#' . static::THIRD_PARTY_ID, $form)) {
      $fields = $formState->getValue('fields');
      foreach ($form['#' . static::THIRD_PARTY_ID] as $widgetName => $widgetData) {
        static::me()
          ->updateWidget($formState, $widgetName, $fields[$widgetName]);
      }
    }
  }

  /**
   * Update the widget data.
   *
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The formstate.
   * @param string $widgetId
   *   The widget Id.
   * @param array $formData
   *   The formData.
   */
  public function updateWidget(FormStateInterface $formState, $widgetId, array $formData) {
    $display = $this->getDisplayFromFormState($formState);
    $widgets = $display->getThirdPartySettings(static::THIRD_PARTY_ID);

    // On actualise les valeurs.
    if (array_key_exists($widgetId, $widgets)) {
      foreach ($widgets[$widgetId] as $key => $value) {
        if (array_key_exists($key, $formData)) {
          $widgets[$widgetId][$key] = $formData[$key];
        }
      }

      $widgets = array_filter($widgets);

      $display->setThirdPartySetting(static::THIRD_PARTY_ID, $widgetId, $widgets[$widgetId]);
      $display->save();
    }
  }

  /**
   * Récupération du display depuis le formulaire.
   *
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The formstate.
   *
   * @return Drupal\Core\Entity\Display\EntityFormDisplayInterface
   *   The entity form display.
   */
  protected function getDisplayFromFormState(FormStateInterface $formState) {
    $callback_object = $formState->getBuildInfo()['callback_object'];
    if (!$callback_object instanceof EntityDisplayFormBase) {
      throw new InvalidArgumentException('Unkown callback object.');
    }

    // Récupération du display.
    return $callback_object->getEntity();
  }

  /**
   * Add the widget in table form.
   *
   * @param array $table
   *   The table.
   * @param string $id
   *   The id.
   * @param array $widget
   *   The widget.
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The formstate.
   * @param mixed $params
   *   The params.
   * @param \Drupal\Core\Entity\Display\EntityFormDisplayInterface $display
   *   The display.
   */
  protected function addWidgetConfig(array &$table, $id, array $widget, FormStateInterface $formState, $params, EntityFormDisplayInterface $display) {
    // A group cannot be selected as its own parent.
    $parent_options = $table['#parent_options'];
    unset($parent_options[$id]);

    $table[$id] = [
      '#attributes'      => ['class' => ['draggable'], 'id' => $id],
      '#row_type'        => 'group',
      '#js_settings'     => ['rowHandler' => 'group'],
      '#region_callback' => get_called_class() . '::regionCallback',
      'human_name'       => [
        '#markup' => Html::escape($this->t('@id', ['@id' => $widget[static::FIELD_ID]])),
        '#prefix' => '<span class="group-label">',
        '#suffix' => '</span>',
      ],
      'weight'           => [
        '#type'          => 'textfield',
        '#default_value' => $widget['weight'],
        '#size'          => 3,
        '#attributes'    => ['class' => ['field-weight']],
      ],
      'parent_wrapper'   => [
        'parent'      => [
          '#type'          => 'select',
          '#options'       => $parent_options,
          '#empty_value'   => '',
          '#default_value' => isset($params->parents[$id]) ? $params->parents[$id] : '',
          '#attributes'    => ['class' => ['field-parent']],
          '#parents'       => ['fields', $id, 'parent'],
        ],
        'hidden_name' => [
          '#type'          => 'hidden',
          '#default_value' => $id,
          '#attributes'    => ['class' => ['field-name']],
        ],
      ],
    ];

    $table[$id]['format'] = FormHelperWidgetAddForm::getFields($widget);

    $base_button = [
      '#submit'     => [
        [$formState->getBuildInfo()['callback_object'], 'multistepSubmit'],
      ],
      '#ajax'       => [
        'callback' => [
          $formState->getBuildInfo()['callback_object'],
          'multistepAjax',
        ],
        'wrapper'  => 'field-display-overview-wrapper',
        'effect'   => 'fade',
      ],
      '#field_name' => $id,
    ];

    if ($formState->get('plugin_settings_edit') == $id) {
      $table[$id]['format']['#cell_attributes'] = ['colspan' => 2];
      $table[$id]['format']['format_settings'] = [
        '#type'       => 'container',
        '#attributes' => ['class' => ['field-plugin-settings-edit-form']],
        '#parents'    => ['fields', $id, 'settings_edit_form'],
        '#weight'     => -5,
        'label'       => [
          '#markup' => $this->t('Field group format:') . ' <span class="formatter-name">' . '</span>',
        ],
        // Create a settings form where hooks can pick in.
        'settings'    => [],
        'actions'     => [
          '#type'           => 'actions',
          'save_settings'   => $base_button + [
            '#type'  => 'submit',
            '#name'  => FormHelperWidgetManager::me()
              ->slugify($id) . '_plugin_settings_update',
            '#value' => $this->t('Update'),
            '#op'    => 'update',
          ],
          'cancel_settings' => $base_button + [
            '#type'                    => 'submit',
            '#name'                    => $id . '_plugin_settings_cancel',
            '#value'                   => $this->t('Cancel'),
            '#op'                      => 'cancel',
            // Do not check errors for the 'Cancel' button.
            '#limit_validation_errors' => [],
          ],
        ],
      ];
      $table[$id]['#attributes']['class'][] = 'field-formatter-settings-editing';
      $table[$id]['format']['type']['#attributes']['class'] = ['visually-hidden'];
    }
    else {
      // After saving, the settings are updated here aswell. First we create
      // the element for the table cell.
      $table[$id]['settings_summary'] = ['#markup' => ''];

      // Add the configure button.
      $table[$id]['settings_edit'] = $base_button + [
        '#type'                    => 'image_button',
        '#name'                    => $id . '_group_settings_edit',
        '#src'                     => 'core/misc/icons/787878/cog.svg',
        '#attributes'              => [
          'class' => ['field-plugin-settings-edit'],
          'alt'   => $this->t('Edit'),
        ],
        '#op'                      => 'edit',
        // Do not check errors for the 'Edit' button, but make sure we get
        // the value of the 'plugin type' select.
        '#limit_validation_errors' => [['fields', $id, 'type']],
        '#prefix'                  => '<div class="field-plugin-settings-edit-wrapper">',
        '#suffix'                  => '</div>',
      ];

      if ($delete_route = $this->getDeleteRoute($id, $display)) {
        $table[$id]['settings_edit']['#suffix'] .= Link::fromTextAndUrl($this->t('delete'), $delete_route)
          ->toString();
      }
    }
  }

  /**
   * Region Callback.
   *
   * @param array $row
   *   The row.
   *
   * @return string
   *   The string.
   */
  public static function regionCallback(array $row) {
    switch ($row['#row_type']) {
      case 'group':
        return 'content';
    }
  }

  /**
   * Retourne la route de suppression.
   *
   * @param string $widgetId
   *   THe widget ID.
   * @param \Drupal\Core\Entity\Display\EntityFormDisplayInterface $display
   *   The display.
   *
   * @return null|Url
   *   The url of the delete route.
   */
  public function getDeleteRoute($widgetId, EntityFormDisplayInterface $display) {
    $entityTypeId = $display->getTargetEntityTypeId();
    $entityType = $this->entityTypeManager->getDefinition($entityTypeId);

    if ($entityType->get('field_ui_base_route')) {

      $parameters = [
        'entityTypeId' => $entityTypeId,
        'mode'         => $display->getMode(),
        'bundle'       => $display->getTargetBundle(),
        'widgetId'     => $widgetId,
      ];

      $routeName = 'field_ui.form_helper_widget_delete_' . $entityType->id();
      return new Url($routeName, $parameters);
    }

    return NULL;
  }

  /**
   * Delete the widget.
   *
   * @param string $entityType
   *   The entityType.
   * @param string $bundle
   *   The entity bundle.
   * @param string $mode
   *   The form mode.
   * @param string $widgetId
   *   THe widget id.
   */
  public function deleteWidget($entityType, $bundle, $mode, $widgetId) {
    if ($display = EntityFormDisplay::load($entityType . '.' . $bundle . '.' . $mode)) {
      $display->unsetThirdPartySetting(static::THIRD_PARTY_ID, $widgetId);
      $display->save();

      $this->messenger
        ->addMessage($this->t('Le widget a correctement été supprimé'));
    }
    else {
      $this->messenger->addMessage($this->t("Une erreur s'est produite"));
    }
  }

  /**
   * Clean les thirds party settings.
   *
   * @param string $entityType
   *   The entitytpe id.
   * @param string $bundle
   *   THe entity bundle.
   * @param string $mode
   *   The display mode.
   */
  public function cleanThirdParty($entityType, $bundle, $mode) {
    if ($display = EntityFormDisplay::load($entityType . '.' . $bundle . '.' . $mode)) {
      $widgets = $display->getThirdPartySettings(static::THIRD_PARTY_ID);
      foreach ($widgets as $widgetId => $widgetData) {
        if (is_null($widgetId) || is_null($widgetData)) {
          $display->unsetThirdPartySetting(static::THIRD_PARTY_ID, $widgetId);
        }
      }

      $display->save();
    }
  }

}
